# -*- coding: utf-8 -*-
"""
caching local containers package.
"""

from pyrin.packaging.base import Package


class CachingLocalContainersPackage(Package):
    """
    caching local containers package class.
    """

    NAME = __name__
