# -*- coding: utf-8 -*-
"""
cli core template package.
"""

from pyrin.packaging.base import Package


class CLICoreTemplatePackage(Package):
    """
    cli core template package class.
    """

    NAME = __name__
