# -*- coding: utf-8 -*-
"""
orm sql schema package.
"""

from pyrin.packaging.base import Package


class ORMSQLSchemaPackage(Package):
    """
    orm sql schema package class.
    """

    NAME = __name__
