# -*- coding: utf-8 -*-
"""
response status package.
"""

from pyrin.packaging.base import Package


class ResponseStatusPackage(Package):
    """
    response status package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'processor.response.status.component'
