# -*- coding: utf-8 -*-
"""
orm sql extractor package.
"""

from pyrin.packaging.base import Package


class ORMSQLExtractorPackage(Package):
    """
    orm sql extractor package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'database.orm.sql.extractor.component'
